TIS_LOGGING_FULL = "full"
TIS_LOGGING_COMPACT = "compact"
TIS_LOGGING_DIFF = "diff"
TIS_LOGGING_NONE = "none"
VALID_TIS_LOGGING_FORMATS = [TIS_LOGGING_FULL, TIS_LOGGING_COMPACT, TIS_LOGGING_NONE]
VALID_TIS_UPDATE_FORMATS = [TIS_LOGGING_FULL, TIS_LOGGING_DIFF]
TIS_LOGGING_AUTO = "auto"
